#include "testlib.h"

#include <cassert>
using namespace std;

// This checker validates that at each step the graph is a tree
// by using the offline dynamic connectivity in approach with in
// $O(n \log^2 n)$ time with $O(n)$ used memory.

const int N = 1 << 19;

// The total number of operations.
int k;

// old_edges[i] - the edge that is alive on [0, i]
// new_edges[i] - the edge that is alive on [i, +inf]
pair<int, int> old_edges[N], new_edges[N];

// DSU with rollback operation support (only the rank heuristic is used).
int par[N];
int rnk[N];

void init_DSU(int n) {
    for (int i = 1; i <= n; i++)
        par[i] = i, rnk[i] = 0;
}

// Stack of performed changes in DSU structure.
int tail = 0;
pair<int*, int> stack[2 * N];

void assign(int& x, int y) {
    stack[tail++] = make_pair(&x, x);
    x = y;
}

// Rolblack the DSU state to the state by the given timestamp.
void rollback(int timestamp) {
    while (tail > timestamp)
        --tail, *stack[tail].first = stack[tail].second;
}

bool merge(int a, int b) {
    if (a < 0 || b < 0)
        return true;
    while (par[a] != a)
        a = par[a];
    while (par[b] != b)
        b = par[b];
    if (a == b)
        return false;
    if (rnk[a] > rnk[b])
        swap(a, b);
    if (rnk[a] == rnk[b])
        assign(rnk[b], rnk[b] + 1);
    assign(par[a], b);
    return true;
}

// Add the given segment of edges in the given array to the DSU
// and return true in case of success and false in case if the
// cycle is formed.
bool process_edges(pair<int, int>* edges, int l, int r) {
    // We are interested in moments of time in the range [0, k]
    for (int i = max(0, l); i <= min(k, r); i++) {
        if (edges[i].first != -1)
            if (!merge(edges[i].first, edges[i].second))
                return false;
    }
    return true;
}

// son_type = 
// * -1 if left son of its parent (we should handle new edges)
// *  0 if root
// *  1 if right son of its parent (we should handle old edges)
void check(int l, int r, int son_type) {
    int timestamp = tail;
    if ((son_type == 1  && !process_edges(new_edges, 2 * l - r, l)) ||
        (son_type == -1 && !process_edges(old_edges, r, 2 * r - l))) {
            quitf(_wa, "There is a cycle in the graph after operation %d (zero-indexed)", l - 1);
    }
    if (l < r) {
        check(l, (l + r) / 2, -1);
        check((l + r) / 2 + 1, r, 1);
    }
    rollback(timestamp);
}

int main(int argc, char* argv[]) {
    registerTestlibCmd(argc, argv);
    int n = inf.readInt();
    init_DSU(n);

    set<pair<int, int> > old_tree, new_tree;
    for (int i = 0; i < n - 1; i++) {
        int a, b;
        a = inf.readInt();
        b = inf.readInt();
        if (a > b)
            swap(a, b);
        old_tree.insert(make_pair(a, b));
    }
    int common_edges = 0;
    for (int i = 0; i < n - 1; i++) {
        int a, b;
        a = inf.readInt();
        b = inf.readInt();
        if (a > b)
            swap(a, b);
        new_tree.insert(make_pair(a, b));
        if (old_tree.find(make_pair(a, b)) != old_tree.end()) {
            common_edges++;
            // This edge will always belong to the graph so we initially add them to the DSU.
            assert(merge(a, b));
        }
    }
    int expected_operations = n - 1 - common_edges;
    
    k = ouf.readInt(-1, n - 1, "number of operations");
    if (k == -1)
        quitf(_wa, "Pretty naive :) contestant printed -1");
    if (k != expected_operations)
        quitf(_wa, "Contestant is going to perform %d operations, but we are sure that there will be %d of them", k, expected_operations);
   
    set<pair<int, int> > current_tree = old_tree;
    
    // Read operations and check that the operations indeed transform the
    // set of the first tree edges to the set of the second tree edges
    // without removing unexisting edges or adding duplicating edges.
    old_edges[k] = new_edges[0] = make_pair(-1, -1);
    for (int i = 0; i < k; i++) {
        int a, b, c, d;
        a = ouf.readInt(1, n, "a[" + vtos(i) + "]");
        b = ouf.readInt(1, n, "b[" + vtos(i) + "]");
        c = ouf.readInt(1, n, "c[" + vtos(i) + "]");
        d = ouf.readInt(1, n, "d[" + vtos(i) + "]");
        if (a > b)
            swap(a, b);
        if (c > d)
            swap(c, d);
        if (current_tree.find(make_pair(a, b)) == current_tree.end())
            quitf(_wa, "Operation %d: there is no edge (%d, %d) in the current tree", i, a, b);
        current_tree.erase(make_pair(a, b));
        if (current_tree.find(make_pair(c, d)) != current_tree.end())
            quitf(_wa, "Operation %d: there is already an edge (%d, %d) in the current tree", i, c, d);
        current_tree.insert(make_pair(c, d));
        
        old_edges[i] = make_pair(a, b);
        new_edges[i + 1] = make_pair(c, d);
    }
    for (set<pair<int, int> >::iterator it = new_tree.begin(); it != new_tree.end(); it++) {
        if (current_tree.find(*it) == current_tree.end()) {
            quitf(_wa, "Edge (%d, %d) doesn't belong to the resulting tree", it->first, it->second);
        }
    }

    // Check connectivity at each step.
    int pw2 = 1;
    while (pw2 - 1 < k)
        pw2 <<= 1;
    check(0, pw2 - 1, 0);

    quitf(_ok, "%d operations required", k);
}
